"""This module handles requests for the network info page."""

from __future__ import annotations

import io
import os
import subprocess
from typing import Any, TYPE_CHECKING, Dict

import configparser
import qrcode
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.shortcuts import render, redirect

from core import util
from core.state_handler import Stateful
from main import settings

if TYPE_CHECKING:
    from core.base import Base


class NetworkInfo(Stateful):
    """This class handles requests on the /network_info page."""

    def __init__(self, base: "Base"):
        self.base = base

    def state_dict(self) -> Dict[str, Any]:
        state_dict = self.base.state_dict()
        return state_dict

    def _qr_path(self, data) -> str:
        # from https://github.com/lincolnloop/python-qrcode/blob/master/qrcode/console_scripts.py
        module = "qrcode.image.svg.SvgPathImage"
        module, name = module.rsplit(".", 1)
        imp = __import__(module, {}, [], [name])
        SvgPathImage = getattr(imp, name)

        qr = qrcode.QRCode()
        qr.add_data(data)
        img = qr.make_image(image_factory=SvgPathImage)
        with io.BytesIO() as stream:
            img.save(stream)
            svg = stream.getvalue().decode()
        tag = svg.split("\n")[1]
        return tag

    def index(self, request: WSGIRequest) -> HttpResponse:
        """Renders the /network_info page. Only admin is allowed to see this page."""
        if not self.base.user_manager.is_admin(request.user):
            return redirect("login")
        context = self.base.context(request)

        # hotspot information
        context["hotspot_enabled"] = False
        try:
            if subprocess.call(["/usr/local/sbin/raveberry/hotspot_enabled"]) != 0:
                # Hotspot is enabled, show its info as well
                context["hotspot_enabled"] = True

                config = configparser.ConfigParser()
                config.read(os.path.join(settings.BASE_DIR, "config/raveberry.ini"))
                ssid = config.get("Hotspot", "hotspot_ssid")
                password = config.get("Hotspot", "hotspot_password")

                device = util.get_devices()[-1]
                ip = util.ip_of_device(device)
                url = f"http://{ip}/"

                context["hotspot_ssid"] = ssid
                context["hotspot_password"] = password
                context["hotspot_wifi_qr"] = self._qr_path(
                    f"WIFI:S:{ssid};T:WPA;P:{password};;"
                )
                context["hotspot_url"] = url
                context["hotspot_url_qr"] = self._qr_path(url)
                context["hotspot_ip"] = ip
        except FileNotFoundError:
            # hotspot was not configured
            pass

        # connected network information
        ssid = None
        password = None
        try:
            ssid = subprocess.check_output(
                "/sbin/iwgetid --raw".split(), universal_newlines=True
            )[:-1]
            wifi_active = True
        except subprocess.CalledProcessError:
            wifi_active = False

        if wifi_active:
            try:
                password = subprocess.check_output(
                    ["sudo", "/usr/local/sbin/raveberry/password_for_ssid", ssid],
                    universal_newlines=True,
                )
            except subprocess.CalledProcessError:
                pass

        device = util.get_devices()[0]
        ip = util.ip_of_device(device)

        wifi_qr = self._qr_path(f"WIFI:S:{ssid};T:WPA;P:{password};;")
        raveberry_url = f"http://{ip}/"
        raveberry_qr = self._qr_path(raveberry_url)

        if wifi_active:
            context["ssid"] = ssid
            context["password"] = password
            context["wifi_qr"] = wifi_qr
        else:
            context["ssid"] = None
            context["password"] = None
            context["wifi_qr"] = None
        context["url"] = raveberry_url
        context["url_qr"] = raveberry_qr
        context["ip"] = ip
        return render(request, "network_info.html", context)



import logging
import time
from typing import Optional, Type, List

from django.conf import settings
from django.db import transaction
from django.db.models.expressions import F

from core.models import (
    ArchivedPlaylist,
    PlaylistEntry,
    ArchivedPlaylistQuery,
    RequestLog,
)
from core.musiq import song_utils as song_utils
from core.musiq.music_provider import MusicProvider, ProviderError
from core.musiq.song_provider import SongProvider


class PlaylistProvider(MusicProvider):
    """The base class for playlist providers."""

    @staticmethod
    def create(
        musiq: "Musiq", query: Optional[str] = None, key: Optional[int] = None,
    ) -> "PlaylistProvider":
        """Factory method to create a playlist provider.
        Both query and key need to be specified.
        Detects the type of provider needed and returns one of corresponding type."""
        if query is None:
            logging.error(
                "archived playlist requested but no query given (key %s)", key
            )
            raise ValueError
        if key is None:
            logging.error("archived playlist requested but no key given")
            raise ValueError
        try:
            archived_playlist = ArchivedPlaylist.objects.get(id=key)
        except ArchivedPlaylist.DoesNotExist:
            logging.error("archived song requested for nonexistent key %s", key)
            raise ValueError

        playlist_type = song_utils.determine_playlist_type(archived_playlist)
        provider_class: Optional[Type[PlaylistProvider]] = None
        if playlist_type == "local":
            from core.musiq.localdrive import LocalPlaylistProvider

            provider_class = LocalPlaylistProvider
        elif playlist_type == "youtube":
            from core.musiq.youtube import YoutubePlaylistProvider

            provider_class = YoutubePlaylistProvider
        elif playlist_type == "spotify":
            from core.musiq.spotify import SpotifyPlaylistProvider

            provider_class = SpotifyPlaylistProvider
        elif playlist_type == "soundcloud":
            from core.musiq.soundcloud import SoundcloudPlaylistProvider

            provider_class = SoundcloudPlaylistProvider
        if not provider_class:
            raise NotImplementedError(f"No provider for given playlist: {query}, {key}")
        provider = provider_class(musiq, query, key)
        return provider

    @staticmethod
    def get_id_from_external_url(url: str) -> Optional[str]:
        """Constructs and returns the external id based on the given url."""
        raise NotImplementedError()

    def __init__(
        self, musiq: "Musiq", query: Optional[str], key: Optional[int]
    ) -> None:
        super().__init__(musiq, query, key)
        self.ok_message = "queueing playlist"
        self.title: Optional[str] = None
        self.urls: List[str] = []

    def check_cached(self) -> bool:
        if self.key is not None:
            archived_playlist = ArchivedPlaylist.objects.get(id=self.key)
        else:
            assert self.query is not None
            try:
                list_id = self.get_id_from_external_url(self.query)
                archived_playlist = ArchivedPlaylist.objects.get(list_id=list_id)
            except (KeyError, ArchivedPlaylist.DoesNotExist):
                return False
        self.id = archived_playlist.list_id
        self.key = archived_playlist.id
        self.urls = [entry.url for entry in archived_playlist.entries.all()]
        return True

    def search_id(self) -> Optional[str]:
        """Fetches the id of this playlist from the internet and returns it."""
        raise NotImplementedError()

    def check_available(self) -> bool:
        if self.id is not None:
            return True
        assert self.query
        list_id = self.get_id_from_external_url(self.query)
        if list_id is None:
            list_id = self.search_id()
        if list_id is None:
            return False
        self.id = list_id
        return True

    def is_radio(self) -> bool:
        """Returns whether this playlist is a radio.
        A radio as a playlist that was created for a given song.
        The result can be different if called another time for the same song."""
        raise NotImplementedError()

    def fetch_metadata(self) -> bool:
        """Fetches the title and list of songs for this playlist from the internet."""
        raise NotImplementedError()

    def enqueue_placeholder(self, manually_requested) -> None:
        # Playlists have no placeholder representation.
        pass

    def remove_placeholder(self) -> None:
        pass

    def make_available(self) -> bool:
        queryset = ArchivedPlaylist.objects.filter(list_id=self.id)
        if not self.is_radio() and queryset.exists():
            archived_playlist = queryset.get()
            self.key = archived_playlist.id
            self.urls = [entry.url for entry in archived_playlist.entries.all()]
        else:
            if not self.fetch_metadata():
                return False
        return True

    def persist(self, request_ip: str, archive: bool = True) -> None:
        if self.is_radio():
            return

        assert self.id
        if self.title is None:
            logging.warning("Persisting a playlist with no title (id %s)", self.id)
            self.title = ""

        with transaction.atomic():
            queryset = ArchivedPlaylist.objects.filter(list_id=self.id)
            if queryset.count() == 0:
                initial_counter = 1 if archive else 0
                archived_playlist = ArchivedPlaylist.objects.create(
                    list_id=self.id, title=self.title, counter=initial_counter
                )
                for index, url in enumerate(self.urls):
                    PlaylistEntry.objects.create(
                        playlist=archived_playlist, index=index, url=url,
                    )
            else:
                if archive:
                    queryset.update(counter=F("counter") + 1)
                archived_playlist = queryset.get()

        if archive:
            ArchivedPlaylistQuery.objects.get_or_create(
                playlist=archived_playlist, query=self.query
            )

        if self.musiq.base.settings.basic.logging_enabled and request_ip:
            RequestLog.objects.create(playlist=archived_playlist, address=request_ip)

    def enqueue(self) -> None:
        for index, external_url in enumerate(self.urls):
            if index == self.musiq.base.settings.basic.max_playlist_items:
                break
            # request every url in the playlist as their own url
            song_provider = SongProvider.create(self.musiq, external_url=external_url)

            try:
                song_provider.request("", archive=False, manually_requested=False)
            except ProviderError:
                continue

            if settings.DEBUG:
                # the sqlite database has problems if songs are pushed very fast
                # while a new song is taken from the queue. Add a delay to mitigate.
                time.sleep(1)
